@@ -39,7 +39,9 @@
 #include "qt3dpresentationnodeinstance.h"
 
 #include "quickitemnodeinstance.h"
+#include "quick3dmaterialnodeinstance.h"
 #include "quick3dnodeinstance.h"
+#include "quick3dtexturenodeinstance.h"
 
 #include "nodeinstanceserver.h"
 #include "instancecontainer.h"
@@ -119,6 +121,11 @@ QImage ServerNodeInstance::renderPreviewImage(const QSize &previewImageSize) con
     return m_nodeInstance->renderPreviewImage(previewImageSize);
 }
 
+QSharedPointer<QQuickItemGrabResult> ServerNodeInstance::createGrabResult() const
+{
+    return m_nodeInstance->createGrabResult();
+}
+
 bool ServerNodeInstance::isRootNodeInstance() const
 {
     return isValid() && m_nodeInstance->isRootNodeInstance();
@@ -129,6 +136,13 @@ bool ServerNodeInstance::isSubclassOf(QObject *object, const QByteArray &superTy
     return  Internal::QmlPrivateGate::isSubclassOf(object, superTypeName);
 }
 
+QRectF ServerNodeInstance::effectAdjustedBoundingRect(QQuickItem *item)
+{
+    if (item)
+        return item->boundingRect().adjusted(-40, -40, 40, 40);
+    return {};
+}
+
 void ServerNodeInstance::setModifiedFlag(bool b)
 {
     m_nodeInstance->setModifiedFlag(b);
@@ -141,7 +155,17 @@ void ServerNodeInstance::setNodeSource(const QString &source)
 
 bool ServerNodeInstance::holdsGraphical() const
 {
-    return m_nodeInstance->isQuickItem();
+    return m_nodeInstance->isRenderable();
+}
+
+bool ServerNodeInstance::isComponentWrap() const
+{
+    return m_nodeInstance->isComponentWrap();
+}
+
+QQuickItem *ServerNodeInstance::contentItem() const
+{
+    return m_nodeInstance->contentItem();
 }
 
 void ServerNodeInstance::updateDirtyNodeRecursive()
@@ -176,8 +200,12 @@ Internal::ObjectNodeInstance::Pointer ServerNodeInstance::createInstance(QObject
         instance = Internal::LayoutNodeInstance::create(objectToBeWrapped);
     else if (isSubclassOf(objectToBeWrapped, "QQuickItem"))
         instance = Internal::QuickItemNodeInstance::create(objectToBeWrapped);
+    else if (isSubclassOf(objectToBeWrapped, "QQuick3DTexture"))
+        instance = Internal::Quick3DTextureNodeInstance::create(objectToBeWrapped);
     else if (isSubclassOf(objectToBeWrapped, "QQuick3DNode"))
         instance = Internal::Quick3DNodeInstance::create(objectToBeWrapped);
+    else if (isSubclassOf(objectToBeWrapped, "QQuick3DMaterial"))
+        instance = Internal::Quick3DMaterialNodeInstance::create(objectToBeWrapped);
     else if (isSubclassOf(objectToBeWrapped, "QQmlComponent"))
         instance = Internal::ComponentNodeInstance::create(objectToBeWrapped);
     else if (objectToBeWrapped->inherits("QQmlAnchorChanges"))
@@ -211,6 +239,35 @@ QString static getErrorString(QQmlEngine *engine, const QString &componentPath)
     return s;
 }
 
+bool isInPathList(const QStringList &pathList, const QString &componentPath)
+{
+    if (componentPath.indexOf("qml/QtQuick/Controls") > 0)
+        return true;
+
+    return std::any_of(pathList.begin(), pathList.end(), [&](auto &&path) {
+        return componentPath.startsWith(path);
+    });
+}
+
+bool canBeCreatedAsPrimitive(const QStringList &pathList,
+                             const InstanceContainer &instanceContainer,
+                             QQmlContext *context,
+                             QObject *&object)
+{
+    if (isInPathList(pathList, instanceContainer.componentPath())) {
+        object = Internal::ObjectNodeInstance::createPrimitive(QString::fromUtf8(
+                                                                   instanceContainer.type()),
+                                                               instanceContainer.majorNumber(),
+                                                               instanceContainer.minorNumber(),
+                                                               context);
+
+        if (object)
+            return true;
+
+    }
+    return false;
+}
+
 ServerNodeInstance ServerNodeInstance::create(NodeInstanceServer *nodeInstanceServer,
                                               const InstanceContainer &instanceContainer,
                                               ComponentWrap componentWrap)
@@ -225,14 +282,19 @@ ServerNodeInstance ServerNodeInstance::create(NodeInstanceServer *nodeInstanceSe
         object = Internal::ObjectNodeInstance::createCustomParserObject(instanceContainer.nodeSource(), nodeInstanceServer->importCode(), nodeInstanceServer->context());
         if (object == nullptr)
             nodeInstanceServer->sendDebugOutput(DebugOutputCommand::ErrorType, QLatin1String("Custom parser object could not be created."), instanceContainer.instanceId());
-    } else if (!instanceContainer.componentPath().isEmpty()) {
+    } else if (!instanceContainer.componentPath().isEmpty()
+               && !canBeCreatedAsPrimitive(nodeInstanceServer->engine()->importPathList(),
+                                instanceContainer, nodeInstanceServer->context(), object)) {
         object = Internal::ObjectNodeInstance::createComponent(instanceContainer.componentPath(), nodeInstanceServer->context());
         if (object == nullptr) {
-            const QString errors = getErrorString(nodeInstanceServer->engine(), instanceContainer.componentPath());
-            const QString message = QString("Component with path %1 could not be created.\n\n").arg(instanceContainer.componentPath());
-            nodeInstanceServer->sendDebugOutput(DebugOutputCommand::ErrorType, message + errors, instanceContainer.instanceId());
+            object = Internal::ObjectNodeInstance::createPrimitive(QString::fromUtf8(instanceContainer.type()), instanceContainer.majorNumber(), instanceContainer.minorNumber(), nodeInstanceServer->context());
+            if (object == nullptr) {
+                const QString errors = getErrorString(nodeInstanceServer->engine(), instanceContainer.componentPath());
+                const QString message = QString("Component with path %1 could not be created.\n\n").arg(instanceContainer.componentPath());
+                nodeInstanceServer->sendDebugOutput(DebugOutputCommand::ErrorType, message + errors, instanceContainer.instanceId());
+            }
         }
-    } else {
+    } else if (!object) {
         object = Internal::ObjectNodeInstance::createPrimitive(QString::fromUtf8(instanceContainer.type()), instanceContainer.majorNumber(), instanceContainer.minorNumber(), nodeInstanceServer->context());
         if (object == nullptr)
             nodeInstanceServer->sendDebugOutput(DebugOutputCommand::ErrorType, QLatin1String("Item could not be created."), instanceContainer.instanceId());
@@ -257,8 +319,17 @@ ServerNodeInstance ServerNodeInstance::create(NodeInstanceServer *nodeInstanceSe
 
     instance.internalInstance()->setInstanceId(instanceContainer.instanceId());
 
+    instance.internalInstance()->setComponentWrap(componentWrap == WrapAsComponent);
+
     instance.internalInstance()->initialize(instance.m_nodeInstance, instanceContainer.metaFlags());
 
+#if QT_VERSION < QT_VERSION_CHECK(6, 2, 1)
+    // Handle hidden state to initialize pickable state
+    nodeInstanceServer->handleInstanceHidden(instance, false, false);
+#else
+    nodeInstanceServer->handlePickTarget(instance);
+#endif
+
     return instance;
 }
 
@@ -335,9 +406,16 @@ void ServerNodeInstance::setPropertyBinding(const PropertyName &name, const QStr
     m_nodeInstance->setPropertyBinding(name, expression);
 }
 
-void ServerNodeInstance::setHideInEditor(bool b)
+void ServerNodeInstance::setHiddenInEditor(bool b)
 {
-    m_nodeInstance->setHideInEditor(b);
+    m_nodeInstance->setHiddenInEditor(b);
+    m_nodeInstance->nodeInstanceServer()->handleInstanceHidden(*this, b, true);
+}
+
+void ServerNodeInstance::setLockedInEditor(bool b)
+{
+    m_nodeInstance->setLockedInEditor(b);
+    m_nodeInstance->nodeInstanceServer()->handleInstanceLocked(*this, b, true);
 }
 
 void ServerNodeInstance::resetProperty(const PropertyName &name)
@@ -458,7 +536,7 @@ QDebug operator <<(QDebug debug, const ServerNodeInstance &instance)
     return debug.space();
 }
 
-uint qHash(const ServerNodeInstance &instance)
+ServerNodeInstance::QHashValueType qHash(const ServerNodeInstance &instance)
 {
     return ::qHash(instance.instanceId());
 }
